unit UniI2c;

{
  IIC-Bus Universal

  Freeware, Copyright 1996, Matthias Weingart;
  comments, bug reports to matthias@penthouse.boerde.de
}

{ universal I2C-businterface }

{ normally each signal have an
   output- and inputwire
  and each signal or line is specified by:
   PortAdr: byte      e.g. $278
   Mask: byte         mostly for one bit e.g. $08
   Inverted: Boolean  logic polarity of signal
}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, StdCtrls;

type
  TAdrType=(tadr8,tadr16);
const
     CEEPROMTimeout=500; {ms timeout until byte written}

{
  type
   TWire = record
       PortAdr:  byte;
       Mask:     byte;
       Inverted: boolean;
   end;

   TSignal = record
        InWire: TWire;
        OutWire: TWire;
   end;
  }

  const
   CI2CDelay      = 5; {5 s Delay == 100kHz }
(*
   {LPT - OUT MASK}
   MSdaOut             = $80; { neg }
   MSclOut             = $08; { not neg, Port+2 }
   {LPT - IN MASK}
   MSdaIn              = $80; { H=True, wegen neg - neg }
   MSclIn              = $08; { H=True}
*)
  const
      CStartbus = $100;
      CReceive  = $200;
      CAck      = $400;
      CStopbus  = $800;

   type
    TLptType= (tlLPT1, tlLPT2, tlLPT3, tlLPT4, tlUserDefined);

type
  TPI2C = class(TComponent)
  private
    { Private-Deklarationen }
   FPort : TLptType;
   FPortAdr : integer;
   FOnPoll : TNotifyEvent;
   FRun : Boolean;
   FLabel: TLabel;
   FDelayCount: LongInt;
   FBeepCount: Integer;

   FInitialPortValue0, FInitialPortValue2: byte;
   FSclInPortAdr : word;
   FSclInMask    : byte;  {Mask=0 means not available}
   FSclInInverted:boolean;
   FSclOutPortAdr :word;
   FSclOutMask    :byte;
   FSclOutInverted:boolean;
   FSdaInPortAdr :word;
   FSdaInMask    :byte;
   FSdaInInverted:boolean;
   FSdaOutPortAdr :word;
   FSdaOutMask    :byte;
   FSdaOutInverted:boolean;
   FChargePumpPortAdr :word;
   FChargePumpMask    :byte;
   FChargePumpInverted:boolean;

 protected
    { Protected-Deklarationen }
   FTimer: TTimer;
   procedure Loaded; override;
   Procedure SetPort(APort : TLptType);
   Procedure SetPortAdr(APortAdr : integer);
   procedure Timer(Sender: TObject);
   procedure SetOutPort0( Value: byte );
   procedure SetOutPort2( Value: byte );
   function GetOutPort0: byte;
   function GetOutPort2: byte;
   procedure SetOutPort0Bit( Mask: byte; bit: Boolean);
   function GetInPortBit( Mask: byte):Boolean;
   Procedure SetOnPoll(AOnPoll : TNotifyEvent);
   procedure SdaHigh;
   procedure SdaLow;
   procedure SclHigh;
   procedure SclLow;
   function  SdaIn: Boolean;
   function  SclIn: Boolean;
   procedure StartBus;
   procedure StopBus;
   function  SendByte( d: byte ): Boolean;
   function  ReceiveByte( last: Boolean ): byte;
  public
    { Public-Deklarationen }
    constructor Create(AOwner:TComponent);override;
    destructor  Destroy; override;
    procedure InitDelay;
    function  CheckBus: Boolean;
    function  Send( dev, sub: byte; count: integer; var data: array of byte ): Boolean;
    function  SendSub2( dev: byte; sub: integer; count: integer; var data: array of byte ): Boolean;
    function  SendMsp( dev: byte; sub: byte; adr: word; value: word): Boolean;
    function  Receive( dev, sub: byte; count: integer; var data: array of byte ): Boolean;
    function  ReceiveSub2( dev:byte; sub: integer; count: integer; var data: array of byte ): Boolean;
    function  ReceiveMsp( dev: byte; sub: byte; adr: word; var value: word ): Boolean;
    procedure Universal( count: integer; var OutData, InData: array of integer );
    function  Available( dev: byte ): Boolean;
    procedure Cancel; {Cancel of clockstretch}
    procedure ReadNvm( AdrType: TAdrType; FAdr: byte; AAdr: Longint; Count: LongInt; var data: array of byte );
    procedure WriteNvm( AdrType: TAdrType; FAdr: byte; AAdr: Longint; Count: LongInt; var data: array of byte );
  published
    { Published-Deklarationen }
   Property Port : TLptType Read FPort Write SetPort;
   Property PortAdr : integer Read FPortAdr Write SetPortAdr;
   Property OnPoll : TNotifyEvent Read FOnPoll Write SetOnPoll;
   Property ErrorLabel : TLabel Read FLabel Write FLabel;
   Property DelayCount: LongInt Read FDelayCount Write FDelayCount;
   Property BeepCount: Integer Read FBeepCount Write FBeepCount;

   Property PortValue0: Byte read GetOutPort0 Write SetOutPort0;
   Property PortValue2: Byte read GetOutPort2 Write SetOutPort2;
   Property InitialPortValue0: Byte Read FInitialPortValue0 Write FInitialPortValue0;
   Property InitialPortValue2: Byte Read FInitialPortValue2 Write FInitialPortValue2;
   Property SclInPortAdr: Word Read FSclInPortAdr Write FSclInPortAdr;
   Property SclInMask:   Byte Read FSclInMask     Write FSclInMask;
   Property SclInInverted: Boolean Read FSclInInverted Write FSclInInverted;
   Property SclOutPortAdr: Word Read FSclOutPortAdr Write FSclOutPortAdr;
   Property SclOutMask:    Byte Read FSclOutMask Write FSclOutMask;
   Property SclOutInverted:Boolean Read FSclOutInverted Write FSclOutInverted;
   Property SdaInPortAdr:  Word Read FSdaInPortAdr Write FSdaInPortAdr;
   Property SdaInMask:     Byte Read FSdaInMask Write FSdaInMask;
   Property SdaInInverted: Boolean Read FSdaInInverted Write FSdaInInverted;
   Property SdaOutPortAdr : Word Read FSdaOutPortAdr Write FSdaOutPortAdr;
   Property SdaOutMask    : Byte Read FSdaOutMask Write FSdaOutMask;
   Property SdaOutInverted: Boolean Read FSdaOutInverted Write FSdaOutInverted;
   Property ChargePumpPortAdr: Word Read FChargePumpPortAdr Write FChargePumpPortAdr;
   Property ChargePumpMask:    Byte Read FChargePumpMask Write FChargePumpMask;
   Property ChargePumpInverted:Boolean Read FChargePumpInverted Write FChargePumpInverted;
  end;

procedure Register;

implementation

{$R-,S-,I-,Q-}
procedure Register;
begin
  RegisterComponents('Hardware', [TPI2C]);
end;

destructor  TPI2C.Destroy;
begin
     inherited Destroy;
     FRun:=False;
     SdaHigh;
     SclHigh;
end;

{ findout LPT-adress}

function GetLptPortFromBios( APort: TLptType ): integer;
begin
     case APort of
          tlLPT1:  Result:=memw[$40:$08];
          tlLPT2:  Result:=memw[$40:$0A];
          tlLPT3:  Result:=memw[$40:$0C];
          tlLPT4:  Result:=memw[$40:$0E];
     end;
end;

function SearchLptAdr( AAdr: integer ): TLptType;
var p: TLptType;
begin
     p:=tlLPT1;
     while ( GetLptPortFromBios(p) <> AAdr ) AND (p<tlUserDefined) do
           inc(p);
     Result:=p;
end;

Procedure TPI2C.SetPort(APort : TLptType);
Begin
   If APort <> FPort Then
      Begin
         FPort := APort;
         if APort<>tlUserDefined then
            FPortAdr := GetLptPortFromBios( APort );
      End;
End;

Procedure TPI2C.SetPortAdr(APortAdr : integer);
Begin
      FPortAdr := APortAdr;
      FPort := SearchLptAdr( APortAdr );
End;

constructor TPI2C.Create(AOwner:TComponent);
var v: word;
begin
   inherited Create(AOwner);
   FPort:=tlLPT1;
   FPortAdr := GetLptPortFromBios( FPort );
   FRun:=False;
   InitDelay;
   FTimer:= TTimer.Create(Self);
   FTimer.Interval := 1;
   FTimer.OnTimer := Timer;
   InitialPortValue0 := $ff;
   InitialPortValue2 := $ff;
(*
   { standardsettings for Philips interface }
   FSclInPortAdr := 1;  {related to FPortAdr}
   FSclInMask    := $08;
   FSclInInverted:= False;
   FSclOutPortAdr := 2;
   FSclOutMask    := $08;
   FSclOutInverted:= False;
   FSdaInPortAdr := 1;
   FSdaInMask    := $80;
   FSdaInInverted:= False;
   FSdaOutPortAdr := 0;
   FSdaOutMask    := $80;
   FSdaOutInverted:= True;
   FChargePumpPortAdr := 0;
   FChargePumpMask    := 0;
   FChargePumpInverted:= False;
 *)
   { standardsettings for mir-tronic LPT smartcard interface }
   FSclInPortAdr := 0;
   FSclInMask    := 00;
   FSclInInverted:= False;

   FSclOutPortAdr := 0;
   FSclOutMask    := $80;
   FSclOutInverted:= False;

   FSdaInPortAdr := 1;
   FSdaInMask    := $80;
   FSdaInInverted:= True;

   FSdaOutPortAdr := 0;
   FSdaOutMask    := $10;
   FSdaOutInverted:= True;

   FChargePumpPortAdr := 0;
   FChargePumpMask    := $01;
   FChargePumpInverted:= False;

   SetOutPort0( InitialPortValue0 );
   SetOutPort2( InitialPortValue2 );
end;

procedure TPI2C.Loaded;
begin
   inherited Loaded;
   if FPort<>tlUserDefined then
      FPortAdr := GetLptPortFromBios( FPort );
   if FPortAdr<>0 then
      System.Port[FPortAdr] := $ff;
end;

procedure TPI2C.Timer(Sender: TObject);
var cp: boolean;
begin
  {toggle chargepump everytime}
  if FChargePumpMask<>0 then
     System.Port[FPortAdr+FChargePumpPortAdr] := System.Port[FPortAdr+FChargePumpPortAdr] XOR FChargePumpMask;
  {call childs}
  if Assigned(FOnPoll) then FOnPoll(Self);
end;

Procedure TPI2C.SetOnPoll(AOnPoll : TNotifyEvent);
Begin
  FOnPoll := AOnPoll;
End;

procedure TPI2C.SetOutPort0( Value: byte );
begin
  if FPortAdr<>0 then
  begin
      System.Port[FPortAdr] := Value;
  end;
end;

procedure TPI2C.SetOutPort2( Value: byte );
begin
  if FPortAdr<>0 then
  begin
      System.Port[FPortAdr+2] := Value;
  end;
end;

function TPI2C.GetOutPort0: byte;
begin
  Result:=$ff;
  if FPortAdr<>0 then
  begin
      Result := System.Port[FPortAdr+0];
  end;
end;

function TPI2C.GetOutPort2: byte;
begin
  Result:=$ff;
  if FPortAdr<>0 then
  begin
      Result := System.Port[FPortAdr+2];
  end;
end;


procedure TPI2C.SetOutPort0Bit( Mask: byte; bit: Boolean);
begin
  if FPortAdr<>0 then
  begin
   if bit then
      System.Port[FPortAdr] := System.Port[FPortAdr] OR Mask
   else
      System.Port[FPortAdr] := System.Port[FPortAdr] AND NOT Mask;
  end;
end;

function  TPI2C.GetInPortBit( Mask: byte):Boolean;
var v: byte;
begin
  Result:=True;
  if FPortAdr<>0 then
  begin
   v:=System.Port[FPortAdr+1];
   Result:=(v AND Mask) = MASK;
  end;
end;

var oi,di: LongInt;

procedure TPI2C.InitDelay;
var o,o1,omax: LongInt;
  procedure DoDelay( oimax: LongInt );
  var dc:integer;
  begin

   oi:=oimax;
   o1:=GetTickCount;
   while oi<>0 do
   begin
     dec(oi);
     di:=1; while di<>0 do dec(di); 
   end;
   o1:=(GetTickCount-o1);

   oi:=oimax;
   o:=GetTickCount;
   while oi<>0 do
   begin
     dec(oi);
     di:=FDelayCount; while di<>0 do dec(di); {ca. 5s}
   end;
   o:=(GetTickCount-o);

  end;
var i:LongInt;   dc:integer; v:Single;
begin
   FDelayCount:=1; omax:=1; o:=0; o1:=0;
   repeat
     omax:=omax*2;
     FDelayCount:= round(sqrt(omax));
     DoDelay(FDelayCount);
   until o>300; {test for at least 300 ms}
   {adjust inner loop to 5s}
   v := (longint(o-o1) * 1000) / CI2CDelay; {ratio outer loop to inner loop}
   FDelayCount := round(omax / v) +1;
(*
   DoDelay( omax div FDelayCount );
   { Testloop for 1 sek. = 200000*5s}
   o:=GetTickCount;
   for i:=0 to 200000 do
   begin
     di:=FDelayCount; while di<>0 do dec(di);
   end;
   o:=GetTickCount-o;
*)
end;

procedure TPI2C.Cancel;
begin
     FRun:=False;
end;

function  TPI2C.SdaIn: Boolean;
begin
     di:=FDelayCount; while di<>0 do dec(di);
     Result := (System.Port[FPortAdr+FSdaInPortAdr] AND FSdaInMask) = FSdaInMask;
     if FSdaInInverted then Result := not Result;
end;

function  TPI2C.SclIn: Boolean;
begin
   di:=FDelayCount; while di<>0 do dec(di);
   if FSclInMask=0 then
      Result:=True
   else
   begin
     Result := (System.Port[FPortAdr+FSclInPortAdr] AND FSclInMask) = FSclInMask;
     if FSclInInverted then Result := not Result;
   end;
end;

procedure TPI2C.SdaHigh;
begin
     if FSdaOutInverted then {set bit to 0}
       System.Port[FPortAdr+FSdaOutPortAdr] := (not FSdaOutMask) and System.Port[FPortAdr+FSdaOutPortAdr]
     else {set bit to 1}
       System.Port[FPortAdr+FSdaOutPortAdr] := FSdaOutMask or System.Port[FPortAdr+FSdaOutPortAdr];
     di:=FDelayCount; while di<>0 do dec(di);
end;

procedure TPI2C.SdaLow;
begin
     if FSdaOutInverted then {set bit to 1}
       System.Port[FPortAdr+FSdaOutPortAdr] := FSdaOutMask or System.Port[FPortAdr+FSdaOutPortAdr]
     else {set bit to 0}
       System.Port[FPortAdr+FSdaOutPortAdr] := (not FSdaOutMask) and System.Port[FPortAdr+FSdaOutPortAdr];
     di:=FDelayCount; while di<>0 do dec(di);
end;

procedure TPI2C.SclHigh;
var i: integer;
begin
     i:=0;
     if FSclOutInverted then {set bit to 0}
       System.Port[FPortAdr+FSclOutPortAdr] := (not FSclOutMask) and System.Port[FPortAdr+FSclOutPortAdr]
     else {set bit to 1}
       System.Port[FPortAdr+FSclOutPortAdr] := FSclOutMask or System.Port[FPortAdr+FSclOutPortAdr];
     while (NOT SclIn) AND FRun do begin
        i:=i+1;
        if i=30000 then
           Application.ProcessMessages;
     end;
end;

procedure TPI2C.SclLow;
begin
     if FSclOutInverted then {set bit to 1}
       System.Port[FPortAdr+FSclOutPortAdr] := FSclOutMask or System.Port[FPortAdr+FSclOutPortAdr]
     else {set bit to 0}
       System.Port[FPortAdr+FSclOutPortAdr] := (not FSclOutMask) and System.Port[FPortAdr+FSclOutPortAdr];
     di:=FDelayCount; while di<>0 do dec(di);
end;

function TPI2C.CheckBus: Boolean;
var d,c: boolean;
    i: integer;
begin
    FRun:=False; {disable ClockStretch}
    SetOutPort0( InitialPortValue0 );
    SetOutPort2( InitialPortValue2 );
    SdaHigh;
    SclHigh;
    d:=SdaIn;
    c:=SclIn;
    Result:= d AND c;
    FRun:=True; {enable ClockStretch}
    if Assigned(FLabel) then
    begin
     if (NOT Result) and Assigned(FLabel) then
     begin
       FLabel.Caption := 'I2C Error:';
       if NOT d then
          FLabel.Caption := ErrorLabel.Caption + 'SDA low ';
       if NOT c then
          FLabel.Caption := ErrorLabel.Caption + 'SCL low ';
       for i:=0 to FBeepCount+1 do
           MessageBeep( word(-1) );
     end
     else
        FLabel.Caption := 'I2C ok.';
    end;
end;

procedure TPI2C.StartBus;
begin
     SdaHigh;
     SclHigh;
     SdaLow;
     SclLow;
end;

procedure TPI2C.StopBus;
begin
     SclLow;
     SdaLow;
     SclHigh;
     SdaHigh;
end;

function TPI2C.SendByte( d: byte ): Boolean;
var i: Integer;
begin
     for i:=0 to 7 do
     begin
        if (d AND $80)=$80 then
           SdaHigh
        else
           SdaLow;
        d:=d shl 1;
        SclHigh;
        SclLow;
     end;
     SdaHigh;
     SclHigh;
     Result := FRun AND NOT SdaIn;
     SclLow;
end;

function TPI2C.Send( dev, sub: byte; count: integer; var data: array of byte ): Boolean;
var i:integer;
begin
 Result:=False;
 if CheckBus then
 begin
  if FPortAdr<>0 then
  begin
     StartBus;
     if SendByte( dev ) then
     begin
        if (sub <> $ff) then
           SendByte( sub );
        for i:=0 to count-1 do
           SendByte( data[i] );
        Result:=True;
     end;
     StopBus;
  end;
 end;
end;

function TPI2C.SendSub2( dev: byte; sub: integer; count: integer; var data: array of byte ): Boolean;
var i:integer;
begin
 Result:=False;
 if CheckBus then
 begin
  if FPortAdr<>0 then
  begin
     StartBus;
     if SendByte( dev ) then
     begin
        SendByte( sub div 256 );
        SendByte( sub );
        for i:=0 to count-1 do
           SendByte( data[i] );
        Result:=True;
     end;
     StopBus;
  end;
 end;
end;

function TPI2C.SendMsp( dev: byte; sub: byte; adr: word; value: word): Boolean;
var i:integer;
begin
 Result:=False;
 if CheckBus then
 begin
  if FPortAdr<>0 then
  begin
     StartBus;
     if SendByte( dev ) then
     begin
        SendByte( sub );
        SendByte( adr div 256 );
        SendByte( adr );
        SendByte( value div 256 );
        SendByte( value );
        Result:=True;
     end;
     StopBus;
  end;
 end;
end;

function TPI2C.ReceiveByte( last: Boolean ): byte;
var i,j: Integer; d: byte;
begin
     SclLow;
     SdaHigh;
     d:=0;
     for i:=0 to 7 do
     begin
        SclHigh;
        d:=d shl 1;
        if (SdaIn) then
           d := d or 1;
        SclLow;
     end;
     if last then
        SdaHigh
     else
        SdaLow;
     SclHigh;
     SclLow;
     SdaHigh;
     Result := d;
end;

function  TPI2C.Receive( dev, sub: byte; count: integer; var data: array of byte ): Boolean;
var i:integer;
begin
 Result:=False;
 if CheckBus then
 begin
  if FPortAdr<>0 then
  begin
     StartBus;
     if sub<>$ff then
     begin {repeated Start}
       if SendByte( dev ) then
       begin
            SendByte( sub );
            StartBus;
       end;
     end;
     if SendByte( dev+1 ) then
     begin
       SdaHigh;
       for i:=0 to count-1 do
         data[i]:=ReceiveByte( i=count-1 );
       Result:=True;
     end;
     StopBus;
  end;
 end;
end;

function  TPI2C.ReceiveSub2( dev:byte; sub: integer; count: integer; var data: array of byte ): Boolean;
var i:integer;
begin
 Result:=False;
 if CheckBus then
 begin
  if FPortAdr<>0 then
  begin
     StartBus;
     if SendByte( dev ) then
     begin
        SendByte( sub div 256 );
        SendByte( sub );
        StartBus;
     end;
     if SendByte( dev+1 ) then
     begin
       SdaHigh;
       for i:=0 to count-1 do
         data[i]:=ReceiveByte( i=count-1 );
       Result:=True;
     end;
     StopBus;
  end;
 end;
end;

function TPI2C.ReceiveMsp( dev: byte; sub: byte; adr: word; var value: word): Boolean;
var i:integer;
begin
 Result:=False;
 if CheckBus then
 begin
  if FPortAdr<>0 then
  begin
     StartBus;
     if SendByte( dev ) then
     begin
        SendByte( sub );
        SendByte( adr div 256 );
        SendByte( adr );
        StartBus;
     end;
     if SendByte( dev+1 ) then
     begin
       SdaHigh;
       value:=256 * ReceiveByte( FALSE );
       value:=value+ReceiveByte( TRUE );
       Result:=True;
     end;
     StopBus;
  end;
 end;
end;

procedure TPI2C.Universal( count: integer; var OutData, InData: array of integer );
var i:integer;
begin
 for i:=0 to count-1 do
       InData[i] := OutData[i];
 if CheckBus then
 begin
  if FPortAdr<>0 then
  begin
     {StartBus;}
     for i:=0 to count-1 do
     begin
       if (OutData[i] and CStartbus)=CStartbus then
          StartBus;
       InData[i] := OutData[i];
       if (OutData[i] and CReceive)=CReceive then
       begin
          SdaHigh;
          InData[i] := InData[i] AND $ff00;
          InData[i] := InData[i] or ReceiveByte( i=count-1 );
          if (i<>count-1) then
             InData[i] := InData[i] or CAck
          else
             InData[i] := InData[i] and NOT CAck;
       end else begin
          if SendByte( outdata[i] and $ff ) then
             InData[i] := InData[i] or CAck
          else
             InData[i] := InData[i] and not CAck;
       end;
       if (OutData[i] and CStopbus)=CStopbus then
          StopBus;
     end;
     {StopBus;}
  end;
 end;
end;

function TPI2C.Available( dev: byte ): Boolean;
begin
 Result:=False;
 if CheckBus then
 begin
  if FPortAdr<>0 then
  begin
     StartBus;
     Result:=SendByte(dev);
     StopBus;
  end;
 end;
end;

procedure TPI2C.ReadNvm( AdrType: TAdrType; FAdr: byte; AAdr: Longint; Count: LongInt; var data: array of byte );
begin
  try
   if FAdr<>$ff then
   begin
     case AdrType of
       tadr8:  Receive(  FAdr+((AAdr div 256) * 2), AAdr mod 256, Count, data );
       tadr16: ReceiveSub2( FAdr+((AAdr div 65536) * 2), AAdr mod 65536, Count, data );
     end;
   end;
  finally
  end;
end;

procedure TPI2C.WriteNvm( AdrType: TAdrType; FAdr: byte; AAdr: Longint; Count: LongInt; var data: array of byte );
var o:LongInt;
begin
  try
   if FAdr<>$ff then
   begin
     case AdrType of
       tadr8:  Send( FAdr+((AAdr div 256) * 2), AAdr mod 256, Count, data );
       tadr16: SendSub2( FAdr+((AAdr div 65536) * 2), AAdr mod 65536, Count, data );
     end;
     o:=GetTickCount+CEEPROMTimeout;
     repeat
       Application.ProcessMessages;
     until Available( FAdr ) or ( o<GetTickCount ) or (not FRun); {wait until write is finished}
   end;
  finally
  end;
end;


end.
